////////////////////////////////////////////////////
// file: robot.cpp                                //
//                                                //
// task: implements the robot class that allows   //
//       to send movement commands to the Robo    //
//       Sapien by producing the same windows     //
//       events that are added to the event       //
//       queue when a user presses the movement   //
//       buttons of the UltraMote Controll.       //
//                                                //
// authors: Team NimbRo,                          //
//          Albert-Ludwigs-University of Freiburg //
//          Germany                               //
//          www.nimbro.net                        //
//          September 2004                        //
//                                                //
// compile: the source-code was programmed and    //
//          compiled with Microsoft's eMbedded    //
//          Visual C++ 3.0 IDE as a part of the   //
//          eMbedded Visual Tools 3.0 - 2002      //
//          Edition. Can be downloaded and used   //
//          for free from www.microsoft.com       //
////////////////////////////////////////////////////


////////////////////////////////////
// Ultra Mote Buttons that we use //
////////////////////////////////////
//  
//    1 2 3 4 5 
//   ---------
// 1| . A . . .
// 2| B C D . .
// 3| . E . . .
// 4| . . . . .
// 5| . . . . .
//
// A: should be programmed with "Move forward"
// B: should be programmed with "Turn left"
// C: should be programmed with "Stop movement"
// D: should be programmed with "Turn right"
// E: should be programmed with "Move backward"


///////////////////
// include files //
///////////////////

// get the class definition
#include "robot.h"

// for using the FindWindow-method
#include "windows.h"
// for using the abs-function
#include "math.h"


/////////////
// methods //
/////////////


/////////////////////////////////////////
// method: robot                       //
// task  : creates a new robot object  //
/////////////////////////////////////////
robot::robot() {

	// start position:upper arm
	currentUpperArmPosition[Left]  = Lower;
	currentUpperArmPosition[Right] = Lower;

	// start position:lower arm
	currentLowerArmPosition[Left]  = Center;
	currentLowerArmPosition[Right] = Center;

} // end-method constructor



/////////////////////////////////////////
// method: ~robot                      //
// task  : deletes a robot object      //
/////////////////////////////////////////
robot::~robot() {
} // end-method destructor


/////////////////////////////////////////
// method: clickUltraMoteCoordinate    //
// task  : clicks at a certain coordi- //
//         nate in the UltraMote       //
//         window.                     //
/////////////////////////////////////////
void robot::clickUltraMoteCoordinate(int _x, int _y) {

	// get a handle to the UltraMote-Window
	HWND hwnd1 = FindWindow(NULL, L"UltraMote");

	// get the child window from the Ultra-Mote-Window that has the point (24,24)
	POINT p; p.x = 24; p.y = 24;
	HWND hwnd2 = ChildWindowFromPoint(hwnd1, p);

	// set the UltraMote-Window as active
	HWND hwnd_active_before = SetActiveWindow(hwnd1);

	// create window event "left mouse button was pressed down at coordinate (x,y)"
	SendMessage(hwnd2, WM_LBUTTONDOWN, 0, MAKELPARAM(_x,_y+24) );

	// create window event "left mouse button was released at coordinate (x,y)"
	SendMessage(hwnd2, WM_LBUTTONUP, 0, MAKELPARAM(_x,_y+24) );

} // end-method clickUltraMoteCoordinate



/////////////////////////////////////////
// method: clickUltraMoteButton        //
// task  : clicks at a certain button  //
//         in the UltraMote window.    //
/////////////////////////////////////////
void robot::clickUltraMoteButton(int _column, int _row) {

	clickUltraMoteCoordinate( (_column-1)*48, (_row-1)*48 );

} // end-method clickUltraMoteButton



/////////////////////////////////////////
// method: Move                        //
// task  : moves the robot forward or  //
//         backward for a certain      //
//         distance.                   //
/////////////////////////////////////////
void robot::Move(int _cm, int _StepType) {

	// RoboSapien movement speed in cm per second
	const float MOVE_SPEED_BIGSTEPS   = 8.0f;
	const float MOVE_SPEED_SMALLSTEPS = 7.0f;

	// for how long do we have to move forward or backward?
	float duration = 0.0f;

	// button-row to press
	int row=1;

	// if _cm is negativ this means we want the robot to move backward
	if (_cm<0) row=3;

	// move with big steps
	if (_StepType==STEPTYPE_BIGSTEPS) {
		duration = _cm/MOVE_SPEED_BIGSTEPS;
		clickUltraMoteButton(2, row);
	}

	// move with small steps
	if (_StepType==STEPTYPE_SMALLSTEPS) {
		duration = _cm/MOVE_SPEED_SMALLSTEPS;
		clickUltraMoteButton(2, row);
		Sleep(100);
		clickUltraMoteButton(2, row);
	}
	
	// let RoboSapien move for a while ...
	Sleep( (unsigned long) (duration * 1000) );

	// stop movement
	Stop();
		
} // end-method Move



/////////////////////////////////////////
// method: Stop                        //
// task  : cancel the last movement or //
//         turning command that was    //
//         send to RoboSapien.         //
/////////////////////////////////////////
void robot::Stop() {

	// click the stop button
	clickUltraMoteButton(2,2);
	Sleep(500);

} // end-method Stop



/////////////////////////////////////////
// method: Turn                        //
// task  : let the robot turn for a    //
//         certain amount of degrees   //
//         to the left or to the       //
//         right.                      //
/////////////////////////////////////////
void robot::Turn(int _deg) {

	// RoboSapien turning speed in degrees per second
	const float TURNING_SPEED_LEFT  = 18.7f;
	const float TURNING_SPEED_RIGHT = 15.0f;
	
	// for how long do we have to turn?
	float duration;
	if (_deg<0)
		duration = -(_deg/TURNING_SPEED_LEFT);
	else
		duration = (_deg/TURNING_SPEED_RIGHT);
	
	// in which column do we have to click?
	int column;

	// turn to the left or to the right?
	if(_deg < 0)
	{
		// turn to the left
		column = 1;
    }
	else
	{
		// turn to the right
		column = 3;
	}				

	// send turn command to RoboSapien
	clickUltraMoteButton(column, 2);

	// let the robot turn for a while ...
	Sleep( (unsigned long) (duration * 1000) );
	
	// robot has turned enough!
	Stop();
				
} // end-method Turn


/////////////////////////////////////////
// method: MoveUpperArm                //
// task  : moves the left or right     //
//         upper arm to a specified    //
//         position.                   //
/////////////////////////////////////////
void robot::MoveUpperArm(UpperArmPosition pos, Side side)
{

	// how many steps do we have to move the side up- or downwards?
	int steps = currentUpperArmPosition[side] - pos;	

	// update upper side position
	currentUpperArmPosition[side] = pos;

	// move side
	for(int i = 1; i <= abs(steps); i++) {

		// side has to be moved downwards
		if (steps>0) {
			clickUltraMoteButton(1,5-side);
		}

		// side has to be moved upwards
		if (steps<0) {
			clickUltraMoteButton(2,5-side);
		}

		// wait 1000 millisecond to complete the motion
		Sleep(1000);

	} // end-for


} // end-method MoveUpperArm


/////////////////////////////////////////
// method: MoveLowerArm                //
// task  : moves the left or right     //
//         lower arm to a specified    //
//         position.                   //
/////////////////////////////////////////
void robot::MoveLowerArm(LowerArmPosition pos, Side side)
{
	// how many steps do we have to move the side up- or downwards?
	int steps = currentLowerArmPosition[side] - pos;	

	// update upper side position
	currentLowerArmPosition[side] = pos;

	// move side
	for(int i = 1; i <= abs(steps); i++) {

		// side has to be moved inwards
		if (steps>0) {
			clickUltraMoteButton(1,1+2*side);
		}

		// side has to be moved outwards
		if (steps<0) {
			clickUltraMoteButton(3,1+2*side);
		}

		// wait 1000 millisecond to complete the motion
		Sleep(1000);

	} // end-for


} // end-method MoveLowerArm


/////////////////////////////////////////
// method: TiltTrunk                   //
// task  : tilts the upper body to the //
//         left or right.              //
/////////////////////////////////////////
void robot::TiltTrunk(Side side)
{
	if(side == Left)
		clickUltraMoteButton(5, 1);
	else
		clickUltraMoteButton(4, 1);

	// wait 1000 milliseconds to complete the motion
	Sleep(1000);

} // end-method TiltTrunk